#tool-feedback {
  list-style-image: url("chrome://feedback/skin/beta-32.png");
}

#tool-feedback[checked="true"] {
  -moz-border-image: url("chrome://browser/skin/images/toggledarkboth-active-64.png") 10 repeat repeat;
}


#feedback-feedback-happy {
  list-style-image: url("chrome://feedback/skin/happy-32.png");
}

#feedback-feedback-sad {
  list-style-image: url("chrome://feedback/skin/sad-32.png");
}
